import numpy as np
import pandas as pd

import gssnng.util as si
import statsmodels.robust.scale

def summed_up(su):
    """
    Just sum up the counts

    :param x: the pandas data frame of ranks, all genes
    :param su: the ranked list of genes *IN* the gene set
    :param sig_len_up: the number of expressed genes matched in the set
    :param norm_method: 'standard or theoretical' # from singscore
    :param score_up: is the rank up or down?  True or False
    """
    # normalise the score for the number of genes in the signature
    score = np.sum(su)
    return(score)


def median_score(su):
    """
    Average Z score

    :param x: the pandas data frame of ranks, all genes
    :param su: the ranked list of genes *IN* the gene set
    :param sig_len_up: the number of expressed genes matched in the set
    :param norm_method: 'standard or theoretical' # from singscore
    :param score_up: is the rank up or down?  True or False
    """
    # normalise the score for the number of genes in the signature
    cnts_med = np.median(su)
    return(cnts_med)


def average_score(su):
    """
    Average Z score

    :param x: the pandas data frame of ranks, all genes
    :param su: the ranked list of genes *IN* the gene set
    :param sig_len_up: the number of expressed genes matched in the set
    :param norm_method: 'standard or theoretical' # from singscore
    :param score_up: is the rank up or down?  True or False
    """
    # normalise the score for the number of genes in the signature
    cnts_mean = np.mean(su)
    return(cnts_mean)


def mean_z(allexprvals, genesetvals):
    """
    Average Z score

    :param exprdat: the pandas data frame of ranks, all genes
    :param su: the ranked list of genes *IN* the gene set
    :param sig_len_up: the number of expressed genes matched in the set
    :param norm_method: 'standard or theoretical' # from singscore
    :param score_up: is the rank up or down?  True or False
    """
    # normalise the score for the number of genes in the signature
    vals_mean = np.mean(allexprvals)
    vals_std = np.std(allexprvals)
    centered = [ (np.abs(x - vals_mean) / vals_std) for x in genesetvals ]
    score = np.mean(centered)
    return(score)


def robust_std(exprdat, su):
    """
    Median of median standardized counts

    :param exprdat: the pandas data frame of ranks, all genes
    :param su: the ranked list of genes *IN* the gene set
    :param sig_len_up: the number of expressed genes matched in the set
    :param norm_method: 'standard or theoretical' # from singscore
    :param score_up: is the rank up or down?  True or False
    """
    # normalise the score for the number of genes in the signature
    cnts_med = np.median(exprdat)
    mad_su = statsmodels.robust.scale.mad(exprdat)
    centered_cnts = [ (np.abs(x - cnts_med) / mad_su) for x in su ]
    score = np.median(centered_cnts)
    return(score)


def rank_biased_overlap(x, exprcol, gs, geneset_genes, limit):
    """
    Rank biased overlap method

    :param x: the pandas data frame of ranks, all genes
    :param su: the ranked list of genes *IN* the gene set
    :param gs: the gene set
    """
    rbo_score = 0.0

    # if undirected, then sort based on centered values
    if gs.mode == '?':
        # center & absolute value ranks
        maxN = np.ceil(len(x.index)/2.0)
        x['undir'] = [ np.abs(xi - maxN) for xi in x[exprcol]]
        exprcol = 'undir'

    # get sorted dataframe
    x_sorted = x.sort_values(by=exprcol, ascending=False)
    y = x_sorted[exprcol]

    for i in range(limit):
        subset = set(y.index[0:(i+1)])
        rbo_score += len(subset.intersection(geneset_genes))

    return( rbo_score )


def singscore(x, su, sig_len, norm_method, gs):
    """
    The singscore method

    :param x: the pandas data frame of ranks, all genes
    :param su: the ranked list of genes *IN* the gene set
    :param sig_len_up: the number of expressed genes matched in the set
    :param norm_method: 'standard or theoretical' # from singscore
    :param score_up: is the rank up or down?  True or False
    :param gs: gene set object
    """
    # normalise the score for the number of genes in the signature
    if gs.mode == '?':
        # center & absolute value ranks
        maxN = np.ceil(len(x.index)/2.0)
        su = [ np.abs(xi - maxN) for xi in su]

    mean_rank = np.mean(su)
    norm_up = si.normalisation(norm_method=norm_method,
                               gs_mode=gs.mode,
                               score=mean_rank,
                               library_len=len(x.index),
                               sig_len=sig_len)

    if gs.mode != '?':
        norm_up = norm_up - 0.5

    return(norm_up)


def ssgsea(x, su, sig_len, omega, gs):
    """
    The ssGSEA method

    :param x: the pandas data frame of ranks, all genes
    :param su: the ranked list of genes *IN* the gene set
    :param sig_len: the number of expressed genes matched in the set
    :param norm_method: 'standard or theoretical' # from singscore
    :param score_up: is the rank up or down?  True or False
    :param gene_set: gene set object

    """
    # The gene expression values for a given sample were rank-normalized

    gene_set = set(gs)

    #first sort by absolute expression value, starting with the highest expressed genes first
    xsorted = x.sort_values(axis=0, ascending=False, inplace=False)
    keys_sorted = xsorted.index.tolist()

    #values representing the ECDF of genes in the geneset
    P_GW_numerator = 0
    P_GW_denominator = 0

    #determining denominator value
    i = 1 #current rank stepping through listing of sorted genes
    for gene in keys_sorted:
        if gene in gene_set:
            P_GW_denominator += i ** omega
        i += 1

    P_GW = lambda : P_GW_numerator / P_GW_denominator

    #values representing the ECDF of genes not in the geneset
    P_NG_numerator = 0
    P_NG_denominator = len(x) - len(gene_set)
    P_NG = lambda : P_NG_numerator / P_NG_denominator

    #integrate different in P_GW and P_NG
    i = 1 #current index in the traversal of sorted genes
    scores = []
    for gene in keys_sorted:
        if gene in gene_set:
            P_GW_numerator += i ** omega
        else:
            P_NG_numerator += 1
        scores.append(P_GW() - P_NG())
        i += 1

    return sum(scores)



def expr_format(x, exprcol, geneset_genes):
    sig_len_up = len(geneset_genes)
    su = []
    for j in geneset_genes:
        if j in x.index:
            su.append(x[exprcol][j])
        else:
            sig_len_up = sig_len_up - 1
    return( (su, sig_len_up) )



def method_selector(gs, x, exprcol, geneset_genes, method, method_params):
    """
    :param gs: the gene set
    :param x: the gene expr data frame
    :param exprcol: the column containing values we'll compute on
    :param geneset_genes: genes in the gene set
    :param method: the method we'll call
    :param method_params: dictionary of method parameters
    :param barcode: cell barcode

    :return: dictionary of results
    """

    (su, sig_len) = expr_format(x, exprcol, geneset_genes)
    exprdat = x[exprcol]

    if method == 'singscore':
        res0 = singscore(exprdat, su, sig_len, method_params['normalization'], gs)

    elif method == 'robust_std':
        res0 = robust_std(exprdat, su)

    elif method == 'summed_up':
        res0 = summed_up(su)

    elif method == 'median_score':
        res0 = median_score(su)

    elif method == 'average_score':
        res0 = average_score(su)

    elif method == 'mean_z':
        res0 = mean_z(exprdat, su)

    elif method == 'rank_biased_overlap':
        res0 = rank_biased_overlap(x, exprcol, gs, geneset_genes, method_params['rbo_depth'])

    elif method == 'ssgsea':
        #x, su, sig_len, omega, gene_set
        res0 = ssgsea(exprdat, su, sig_len, method_params['omega'], geneset_genes)

    else:
        return(np.nan)

    return(res0)


def scorefun(gs,
             x,
             method,
             method_params,
             ranked):
    """
    given a ranked list, produce a score

    :param gs: the gene set
    :param x: the pandas data frame of ranks, all genes
    :param method: the method we'll call
    :param method_params: dictionary of method parameters
    :param ranked: ranked data? True | False

    :return a score
    """

    try:
        if (gs.mode == 'UP') and (ranked == False):
            res0 = method_selector(gs, x, 'counts', gs.genes_up, method, method_params)

        elif (gs.mode == 'DN') and (ranked == False):
            res0 = method_selector(gs, x, 'counts', gs.genes_dn, method, method_params)

        elif (gs.mode == 'BOTH') and (ranked == False):
            res0_up = method_selector(gs, x, 'counts', gs.genes_up, method, method_params)
            res0_dn = method_selector(gs, x, 'counts', gs.genes_dn, method, method_params)
            res0 = (res0_up + res0_dn)

        elif (gs.mode == '?') and (ranked == False):
            res0 = method_selector(gs, x, 'counts', gs.genes_up, method, method_params)

        elif (gs.mode == 'UP') and (ranked == True):
            res0 = method_selector(gs, x, 'uprank', gs.genes_up, method, method_params)

        elif (gs.mode == 'DN') and (ranked == True):
            res0 = method_selector(gs, x, 'dnrank', gs.genes_dn, method, method_params)

        elif (gs.mode == 'BOTH') and (ranked == True):
            res0_up = method_selector(gs, x, 'uprank', gs.genes_up , method, method_params)
            res0_dn = method_selector(gs, x, 'dnrank', gs.genes_dn, method, method_params)
            res0 = (res0_up + res0_dn)

        elif (gs.mode == '?') and (ranked == True):
            res0 = method_selector(gs, x, 'uprank', gs.genes_up, method, method_params)

    except ():
        #res1 = dict(barcode = barcode, name=gs.name, mode=gs.mode, score=np.nan, var=np.nan)
        res0 = np.nan

    return(res0)




import logging

logger = logging.getLogger('gssnng')
logger.setLevel(logging.INFO)
logger.propagate = False
handler = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s %(name)-12s %(levelname)-8s %('
                              'messages)s')
handler.setFormatter(formatter)

if not logger.handlers:
    logger.addHandler(handler)

from . import score_cells
from .score_cells import with_gene_sets



import anndata
import numpy as np
from scipy import sparse
import pandas as pd
import scanpy as sc
from anndata import AnnData
from gssnng.smoothing import nn_smoothing
from gssnng.util import error_checking
from gssnng.score_funs import scorefun
from gssnng.gene_sets import genesets
from typing import Union
from multiprocessing import Pool


def with_gene_sets(
        adata: anndata.AnnData,
        gene_set_file: str,
        groupby: Union[str, list, dict],
        smooth_mode: str,
        recompute_neighbors: int,
        score_method: str,
        method_params: dict,
        samp_neighbors: int,
        ranked: bool,
        cores: int
    ) -> anndata.AnnData:

    """
    gene set scoring (all gene sets in file) with nearest neighbor smoothing of the expression matrix

    Improved single cell scoring by:
    - smoothing the data matrix
        - adding noise to the nearest neighbor smoothing via `samp_neighbors`
    - adding noise to the expression data itself (via noise_trials)

    :param adata: anndata.AnnData containing the cells to be scored
    :param gene_set_file: the gene set file with list of gene sets, gmt, one per line
    :param groupby: either a column label in adata.obs, and all categories taken, or a dict specifies one group.
    :param smooth_mode: `adjacency` or `connectivity`, which representation of the neighborhood graph to use.
        `adjacency` weights all neighbors equally, `connectivity` weights close neighbors more
    :param recompute_neighbors: should neighbors be recomputed within each group, 0 for no, >0 for yes and specifies N
    :param score_method: which scoring method to use
    :param method_params: specific params for each method.
    :param samp_neighbors: number of neighbors to sample
    :param ranked: whether the gene expression counts should be rank ordered
    :param cores: number of parallel processes to work through groupby groups

    :returns: adata with gene set scores in .obs
    """

    noise_trials = 0  ### not used currently

    # our gene set data object list
    gs_obj = genesets(gene_set_file)

    error_checking(adata, samp_neighbors, recompute_neighbors,
                   gs_obj, score_method, ranked, method_params)

    if method_params == None:
        method_params = dict()

    # score each cell with the list of gene sets
    all_scores = _proc_data(adata, gs_obj, groupby, smooth_mode, recompute_neighbors,
                                  score_method, method_params, samp_neighbors,
                                  noise_trials, ranked, cores)
    ## join in new results
    adata.obs = adata.obs.join(all_scores, how='left')

    print("**done**")
    return(adata)


def _smooth_out(adata, samp_neighbors, smooth_mode):
    """
    calculates a neighourhood-smoothed signal of adata.X.
    :param samp_neighbors: how many NN to subsample (hence introducing some noise)
    :param smooth_mode: 'connectivity', 'adjacency', see nn_smoothing()

    Returns a complete AnnData with smooeth adata.X
    """
    if smooth_mode not in ['connectivity', 'adjacency']:
        print("ERROR:  please use smooth mode: `adjacency` or `connectivity`")
        exit()
    smoothed_matrix = nn_smoothing(adata.X, adata, smooth_mode, samp_neighbors)
    # for easier handling with gene names
    # smoothed_adata = AnnData(smoothed_matrix, obs=adata.obs, var=adata.var)
    adata.obsm['X_smooth'] = smoothed_matrix
    return(adata)


def _build_data_list(
        adata,
        groupby,
        cats,
        recompute_neighbors,
        samp_neighbors,
        smooth_mode,
):
    """
    creates the smoothed adata.
    Smoothing proceeds within splits via 'groupby' (i.e. a subset of the true NN graph)
    """
    data_list = []  # list of dicts
    for ci in cats:
        # then for each group
        qi = adata[adata.obs['gssnng_groupby'] == ci]
        if recompute_neighbors > 0:
            print("RECOMPUTING KNN")
            sc.pp.neighbors(qi, n_neighbors=recompute_neighbors)
        # smooth and make an adata, and add to the list
        qi_smoothed = _smooth_out(qi, samp_neighbors, smooth_mode)
        params = (qi_smoothed, ci)
        data_list.append(params)
    # send off for scores
    return(data_list)


def _proc_data(
        adata: anndata.AnnData,
        gs_obj: genesets,
        groupby: Union[str, list, dict],
        smooth_mode: str,
        recompute_neighbors: int,
        score_method: str,
        method_params: dict,
        samp_neighbors: int,
        noise_trials: int,
        ranked: bool,
        cores: int
                     ):
    """
    In many cases, the neighbors should be defined.  If you have mixed clinical endpoints,
    you might want scores for only the early points together, and scores for later points.
    Or T-cell neighbors should maybe only be other T-cells. etc. Or by leiden cluster label.
    By building out the list of groups, it can also be run in parallel.

    :param adata: anndata.AnnData containing the cells to be scored
    :param gs_obj: the gene sets class object
    :param groupby: either a column label in adata.obs, and all categories taken, or a dict specifies one group.
    :param recompute_neighbors: should the neighbors be recomputed within each group?  SLOW SLOW SLOW
    :param score_method: which scoring method to use
    :param method_params: specific params for each method.
    :param samp_neighbors: number of neighbors to sample
    :param noise_trials: number of noisy samples to create, integer
    :param ranked: whether the gene expression counts should be rank ordered
    :param cores: number of parallel processes to work through groupby groups

    :returns: scores in a dict for each cell in a list.
    """
    data_list = []  # list of dicts

    if groupby is None:  # take all cells
        cats = ['cat1']
        adata.obs['gssnng_groupby'] = 'cat1'  # altered to gssnng_groupby

    # here we group by one category
    elif isinstance(groupby, str):
        adata.obs['gssnng_groupby'] = adata.obs[groupby]  #
        cats = set(adata.obs[groupby])  # categories

    # here we groupby an intersection of categories, up to 4
    elif isinstance(groupby, list):
        # probably more readable:
        # glist = [adata.obs[g] for g in groupby ]
        # zipped_cols = zip(*glist)
        if len(groupby) == 2:
            zipped_cols = list(zip(adata.obs[groupby[0]], adata.obs[groupby[1]]))
        elif len(groupby) == 3:
            zipped_cols = list(zip(adata.obs[groupby[0]], adata.obs[groupby[1]], adata.obs[groupby[2]]))
        elif len(groupby) == 4:
            zipped_cols = list(zip(adata.obs[groupby[0]], adata.obs[groupby[1]], adata.obs[groupby[2]], adata.obs[groupby[3]]))
        else:
            raise Exception("ERROR: Number of groups must be 4 or less")
        adata.obs['gssnng_groupby'] = zipped_cols  #
        cats = set(zipped_cols)  # categories

    elif isinstance(groupby, dict):
        # then we want to only score
        return("ERROR: dict not implemented")

    data_list = _build_data_list(adata, groupby, cats, recompute_neighbors, samp_neighbors, smooth_mode)
    # then we can start scoring cells #

    # building up the argument list for the parallel call of _score_all_cells_all_sets
    arglist = []
    for smoothed_adata, groupname in data_list:
        arglist.append(
            (smoothed_adata, gs_obj, score_method, method_params, noise_trials, ranked, groupname)
        )
    # how about lambda x: _score_all_cells_all_sets(x, gs_obj, score_method, method_params, noise_trials, ranked)
    # not sure if that works with parallel, due to pickling. might have to be a proper function
    with Pool(processes=cores) as pool:
        res0 = pool.starmap_async(_score_all_cells_all_sets, arglist).get()
    # we column-bind the vectors into a pandas table
    res1 = pd.concat(res0, axis=0)
    return(res1)


def _get_cell_data(
        smoothed_adata: anndata.AnnData,
        cell_ix: int,
        noise_trials: int,
        method_params: dict,
        ranked: bool
) -> pd.DataFrame:
    """
    the processed expression data for each cell: Pulls out the cells expression vector,
    optinoally adding some noise

    :param smoothed_adata: anndata.AnnData containing the cells to be scored
    :param cell_ix: index of the cell in adata
    :param noise_trials: number of noisy samples to create, integer
    :param method_params: specific params for each method.

    :return: a data frame
    """
    # for each cell, rank the expression
    #gene_mat = smoothed_adata.X[cell_ix]
    gene_mat = (smoothed_adata.obsm['X_smooth'])[cell_ix]
    # then we subset it to only the genes with counts
    _, gdx, _ = sparse.find(gene_mat)

    if gene_mat.ndim == 2:
        df = pd.DataFrame(gene_mat[:, gdx].A.flatten(), index=smoothed_adata.var.index[gdx])  ## ????
    else:
        df = pd.DataFrame(gene_mat[gdx],
                          index=smoothed_adata.var.index[gdx])  ## not sure why it's coming off as an array
    df.columns = ['counts']

    if ('normalization' in method_params) and (method_params['normalization'] == 'average') and (noise_trials > 0):
        # add some noise to gene counts.. create a n numbers of examples
        raise ValueError('not implemented')
        #df_noise = si.add_noise(df, noise_trials, 0.01, 0.99)  ## slow part .. fixed
    else:
        df_noise = df

    if ranked:
        # for right now always ranking genes up #
        df_noise['uprank'] = df_noise.iloc[:,0].rank(method='min', ascending=True)  # up or down
        df_noise['dnrank'] = np.max(df['uprank']) - df['uprank']

    return(df_noise)


def _score_all_cells_all_sets(
        smoothed_adata: AnnData,
        gene_set_obj: genesets,
        score_method: str,
        method_params: dict,
        noise_trials: int,
        ranked: bool,
        group_name: str
        ) -> pd.DataFrame:
    """
    Process cells and score each with a list of gene sets and a method

    :param smoothed_adata: anndata.AnnData containing the cells to be scored
    :param gene_set_obj: the gene sets class object
    :param score_method: which scoring method to use
    :param method_params: specific params for each method.
    :param noise_trials: number of noisy samples to create, integer
    :param ranked: whether the gene expression counts should be rank ordered
    :param group_name: group of cells currently being processed

    :return: list of list of gene set score dictionaries
    """
    print("running " + str(group_name))

    results_list = []
    barcodes = []
    for cell_ix in range(smoothed_adata.shape[0]):  # tqdm.trange(smoothed_adata.shape[0]):  # for each cell ID
        results = dict()
        df_cell = _get_cell_data(smoothed_adata, cell_ix, noise_trials, method_params, ranked)
        barcodes.append(smoothed_adata.obs.index[cell_ix])
        for gs_i in gene_set_obj.set_list:  # for each gene set
            results[gs_i.name] = scorefun(gs_i, df_cell, score_method, method_params, ranked)
        results_list.append(results)
    results_df = pd.DataFrame(results_list, index=barcodes)
    return(results_df)




class geneset:
    """ class object that holds a gene set, can be combination of up and dn"""

    def __init__(self, name: str, info: str, gs_up: list, gs_dn: list, mode='NA'):
        """
        :param name: gene set name
        :param info: gene set info
        :param genes_up: list of up regulated genes
        :param genes_dn:  list of down regulated genes
        :param mode:  what type of geneset, up, dn, or both
        """
        self.name = name
        self.info = info
        if mode == 'UP':
            self.genes_up = gs_up
            self.mode = 'UP'
        elif mode == 'DN':
            self.genes_dn = gs_dn
            self.mode = 'DN'
        elif mode == '?':
            self.genes_up = gs_up
            self.mode = '?'
        else:
            self.genes_up = gs_up
            self.genes_dn = gs_dn
            self.mode = 'BOTH'

    def get_name(self):
        return(self.name)

    def up_gene_size(self):
        return (len(self.genes_up))

    def check_direction(self):
        if self.mode == "MATCHED":
            return("MATCHED")
        if 'UP' in self.name.upper():
            self.mode = 'UP'
            return('UP')
        elif 'DN' in self.name.upper():
            self.mode = 'DN'
            return ('DN')
        else:
            self.mode = 'UP'  # default
            return('default')


class genesets:

    def __init__(self, gmt_file):
        # initially all gene sets are entered as UP
        # then in the cleaning step, they are assigned as UP,DN,or MATCHED

        self.set_list = []  # list of geneset obj
        with open(gmt_file) as fh:
            for line in fh:
                bits = line.split('\t')
                if len(bits) >= 3:
                    self.set_list.append(geneset(name=bits[0], info=bits[1], gs_up=bits[2:], gs_dn=[], mode='?'))
        self.set_list = self.clean_sets()

    def trim_name(self, gs):
        name = gs.name[0: (len(gs.name) - 2)]
        if name[(len(name)-1)] == '.':
            name = name[:name.rfind(".")]
        if name[(len(name)-1)] == '_':
            name = name[:name.rfind("_")]
        return(name)

    def clean_sets(self):
        """
        look for up and dn signifiers
        and join up-dn pairs.
        undefined sets will remain "?", could be up or down
        """
        cleaned_sets = []
        for gs1 in self.set_list:
            di = gs1.check_direction()
            if gs1.mode != 'MATCHED' and di in ['UP', 'DN']: ### may need to be matched
                # then we need to check for a pair
                match_flag = 0
                trimmed_name = self.trim_name(gs1)
                for gs2 in self.set_list:
                    trimmed_name2 = self.trim_name(gs2)
                    if ( (trimmed_name == trimmed_name2) and (gs1.name != gs2.name) and
                            (gs1.mode != "MATCHED") and (gs2.mode != "MATCHED") ):
                        gs1.mode = "MATCHED"
                        gs2.mode = "MATCHED"
                        match_flag = 1
                        if di == 'UP': ### gs1 is UP and gs2 is DN
                            cleaned_sets.append(geneset(name=trimmed_name, info=gs1.info,
                                                             gs_up=gs1.genes_up, gs_dn=gs2.genes_up, mode='BOTH'))
                        else: ### gs1 is down and gs2 is up
                            cleaned_sets.append(geneset(name=trimmed_name, info=gs1.info,
                                                             gs_up=gs2.genes_up, gs_dn=gs1.genes_up, mode='BOTH'))
                        break
                if match_flag == 0: ### didn't match, update direction
                    if di == 'UP':
                        cleaned_sets.append(geneset(name=gs1.name, info=gs1.info, gs_up=gs1.genes_up, gs_dn=[], mode='UP'))
                    else:
                        cleaned_sets.append(geneset(name=gs1.name, info=gs1.info, gs_up=[], gs_dn=gs1.genes_up, mode='DN'))
            # then it's default and added as ?
            elif gs1.mode != "MATCHED":
                cleaned_sets.append(geneset(name=gs1.name, info=gs1.info, gs_up=gs1.genes_up, gs_dn=[], mode='?'))
        return(cleaned_sets)

    def num_genesets(self):
        return(len(self.set_list))

    def get_gs_names(self):
        gs_names = [gs.get_name() for gs in self.set_list]
        return(gs_names)


import numpy as np
from scipy import sparse
import logging


NN_DISTANCE_KEY = 'distances'  # scanpy names in .obsp
NN_CONN_KEY = 'connectivities'

# TODO test: should always sum to 1
# multiplying should leave a "one-vector" still sum to one


def get_smoothing_matrix(adata, mode, add_diag):
    """
    using the nearest neighbor graph in adata.obsp, calculate the smoothing
    matrix S such that S @ X smoothes the signal X over neighbors

    :param adata: anndata.AnnData gene expression matrix to smooth
    :param mode: `adjacency` or `connectivity`, which representation of the neighborhood graph to use.
        `adjacency` weights all neighbors equally, `connectivity` weights close neighbors more
    :param add_diag: should the datapoint itself be consider in the smoothing. This should ==True in almost all cases!

    :return: scipy.sparse sparse matrix with the smoothed expression
    """

    if mode == 'adjacency':
        A = (adata.obsp[NN_DISTANCE_KEY] > 0).astype(int)

        # add the diagnoal, ie. the datapoint itself should be represented
        # in the smoothing!
        assert np.all(A.diagonal() == 0), "diagonal of distance matrix not 0!!"
        if add_diag:
            A = A + sparse.diags(np.ones(A.shape[0]))

        # normalize to sum=1 per  row
        row_scaler = 1 / A.sum(axis=1).A.flatten()

        # multiplying with a diag matrix d on the left multiplys each row i by d_i
        normA = sparse.diags(row_scaler) @ A
        return normA

    # actually works exactly the same; could just switch out the obsp key
    elif mode == 'connectivity':
        A = adata.obsp[NN_CONN_KEY]
        # add the diagnoal, ie. the datapoint itself should be represented
        # in the smoothing! Note that the max connectivity == 1
        assert np.all(A.diagonal() == 0), "diagonal of connectivity matrix not 0!!"
        if add_diag:
            A = A + sparse.diags(np.ones(A.shape[0]))
        # normalize to sum=1 per  row
        row_scaler = 1 / A.sum(axis=1).A.flatten()
        normA = sparse.diags(row_scaler) @ A
        return normA
    else:
        raise ValueError(f'unknown mode {mode}')


def random_mask_a_nn_matrix(X, nn_to_keep):
    """
    for a nearest neighbour matrix (i.e. each row has N entries)
    subsample the neighours (setting some entries per row to 0)

    this is pretty slow, maybe there's a better way...

    :param X: nearest neighbor matrix (sparse)
    :nn_to_keep: out of the N nearest neighbors how many to keep (the others will be set to zero)

    :return: sparse.csr_matrix with randomly subsampled nearest neighbors
    """
    # TODO if X is in csr format, we can quickly subsample!
    # just set some elements of data[intptr[row]:intptr[row+1]] to zero
    # and eliminate zeros

    newrows = []
    newcols = []
    newvals = []
    for i in range(X.shape[0]):
        _, col_ix, vals = sparse.find(X[i])
        col_ix_len = len(col_ix)
        if nn_to_keep > col_ix_len:
            samp_n = col_ix_len
        else:
            samp_n = nn_to_keep  # sometimes there are not exactly the expected number of values.
        rand_ix = np.random.choice(col_ix_len, size=samp_n, replace=False)
        newrows.extend([i]*samp_n)
        newcols.extend(col_ix[rand_ix])
        newvals.extend(vals[rand_ix])

    # TODO choose matrix format as X
    return sparse.csr_matrix((newvals, (newrows, newcols)))


def nn_smoothing(X, adata, mode, samp_neighbors, add_diag=True):
    """
    smooth the expression matrix X (cells x genes) using the neighborhood
    graph stored in adata.

    :param X: data to smooth. (cell x feature) matrix
    :param adata: sc.AnnData, containing the neighborhood graph
    :param samp_neighbors: consider all neighbours or a subsample of size samp_neighbors
    :param add_diag: consider the datapoint itself in smoothing. Should be True in almost all cases

    :return: scipy.sparse matrix with the smoothed signals
    """
    assert X.shape[0] == adata.shape[0], "cell number mismatch"
    logging.info("creating smoothing matrix")
    smoothing_mat = get_smoothing_matrix(adata, mode, add_diag)

    if samp_neighbors > 0:
        # randomly set some edges/neighbours to zero
        # experimental and pretty slow!
        logging.info("creating random mask")
        smoothing_mat = random_mask_a_nn_matrix(smoothing_mat, samp_neighbors)

    smooth_signals = smoothing_mat @ X

    return(smooth_signals)



import matplotlib
import numpy as np

matplotlib.use("Agg")
import numpy
import pandas
import logging
logger = logging.getLogger('gssnng')

def error_checking(
        adata,
        samp_neighbors,
        recompute_neighbors,
        gs_obj,
        score_method,
        ranked,
        method_params
):
    """
    QC on the adata. Need to make sure there's enough neighbors available given the sampling size.

    :param adata: the AnnData object
    :param samp_neighbors: integer, number of neighbors to sample
    """

    if type(method_params) != type(dict()):
        raise Exception('ERROR: please use a dictionary to pass method params')

    if any([xi in adata.obs.columns for xi in gs_obj.get_gs_names()]):
        #raise Exception('ERROR: gene set names in columns of adata.obs, please drop.')
        print("Warning! Dropping gene set names from obs!")
        genesetlist = [x.name for x in gs_obj.set_list]
        adata.obs.drop(columns=genesetlist, inplace=True)

    if 'gssnng_groupby' in adata.obs.columns:
        adata.obs.drop(columns='gssnng_groupby', inplace=True)
        #raise Exception("Error: please drop 'gssnng_groupby' as a column name.")
        print('... and dropping gssnng_groupby column...')

    if ranked == False and score_method == 'singscore':
        raise Exception('ERROR: singscore requires ranked data, set ranked parameter to True')

    if (recompute_neighbors == None) or (recompute_neighbors == 0):
        n_neighbors = adata.uns['neighbors']['params']['n_neighbors'] #[0]# in older AnnData versions need this??
    else:
        n_neighbors = recompute_neighbors

    if n_neighbors < samp_neighbors:
        print('*******')
        print('WARNING: Number of neighbors too low for sampling parameter!')
        print('Please reduce number of neighbor samples or recompute neighbor graph.')
        raise Exception('Samp Neighbors Error')
    else:
        return(True)


def read_gene_sets(filepath):
    txt = open(filepath).read().split('\n')
    gd = dict()
    for line in txt:
        bits = line.split('\t')
        if len(bits) >= 3:
            gd[bits[0]] = bits[2:]
    return(gd)


def add_noise(df, n, noise_low, noise_high):
    #df2 = df.copy()
    # draw all random noise in one call
    dfs = pandas.Series(df.gene_counts)
    randmat = pandas.DataFrame(numpy.random.uniform(noise_low, noise_high, (dfs.size, n) ) )
    randmat.index = df.index
    randmat = randmat.apply(lambda x: x + dfs)
    return(randmat)


def get_conn_dist(q, celli, nn):
    # q is an adata
    # celli is cell i in q
    # nn is the number of neighbors
    jdx = numpy.where(q.obsp['distances'][celli].todense() > 0)[1]
    y = [q.obsp['distances'][celli].todense().tolist()[0][ i ] for i in jdx]
    x = [q.obsp['connectivities'][celli].todense().tolist()[0][ i ] for i in jdx]
    df = pandas.DataFrame({'idx':jdx, 'conn':x, 'dist':y})
    sumconn = sum(df['conn'])
    df['prob'] = [x/sumconn for x in df['conn']]
    return(df)


def to_dense_transpose_list(gene_counts):
    gene_mat = gene_counts.todense().transpose().sum(axis=1)
    gdx = numpy.argwhere(gene_mat > 0)
    return( (gene_mat,  [x[0] for x in gdx] ) )


## From SingScore
def normalisation(norm_method,
                  gs_mode,
                  score,
                  library_len,
                  sig_len):

    """

    :param norm_method: method of normalisation, standard or theoretical
    :param score_list: list of scores will each be normalised
    :param score: average score (average of list)
    :param library_len: length of the library (int)
    :param sig_len: length of the signature used to generate the scores

    :return: a tuple, containing the normalised score (float) and an array of
    each genes score normalised
    """
    try:
        if norm_method == 'standard':
            norm = score / library_len
        elif norm_method == 'theoretical':
            if gs_mode != '?':
                low_bound = (sig_len + 1) / 2
                upper_bound = ((2*library_len) - sig_len + 1)/2
                norm = (score - low_bound) / (upper_bound - low_bound)
            else:
                low_bound = (np.ceil(sig_len/2) + 1) / 2
                upper_bound = (library_len - np.ceil(sig_len/2) + 1) / 2
                norm = (score - low_bound) / (upper_bound - low_bound)

        return norm

    except:
        logger.exception('Normalisation method must be standard or theoretical.')


def normalisation_rank(norm_method, ranks, library_len, sig_len):
    """
    :param norm_method: method of normalisation, standard or theoretical
    :param ranks: a dataframe of ranks
    :param library_len: length of library (int)
    :param sig_len: length of the signature used to generate the dataframe

    :return: a dataframe with normalised ranks for each gene in each sample
    """
    try:
        if norm_method == 'standard':
            ranks = ranks/library_len
        elif norm_method == 'theoretical':
            low_bound = (sig_len + 1) / 2
            upper_bound = ((2 * library_len) - sig_len + 1) / 2
            ranks = (ranks- low_bound)/(upper_bound-low_bound)

        return ranks

    except:
        logger.exception('Normalisation method must be standard or theoretical.')


def genesets_long_to_gmt(file_name_in, gmt_file_out, gs_mode, set_name_column, symbol_column, desc_column=None, header=True):
    """
    :param file_name_in: file path to long formatted gene set file
    :param gmt_file_out: the path and filename for the newly formatted gmt file
    :param gs_mode: gene set scoring mode, 'up', 'dn', or 'both'
    :param set_name_column: integer, indicating the column containing the get set name
    :param symbol_column: integer, indicates column containing the gene symbol or ID
    :param desc_column: integer, indicates column containing the gene set description, if none, gene set name used.
    :param header: boolean, indicates whether the first row in file_name_in contains column names (a header)
    :return: None, no return
    """
    # if you have a two column file with
    # gene set in set_name_column and gene symbol in symbol_column
    # write a new gmt formatted file.
    txt = open(file_name_in, 'r').read().split('\n')
    outtxt = open(gmt_file_out,'w')
    gd = dict()

    if header: # drop the first row
        txt = txt[1:]

    if desc_column is None:
        desc_column = set_name_column

    i = set_name_column
    j = symbol_column
    k = desc_column

    for line in txt:
        if '.tsv' in file_name_in:
            bits = line.strip().split('\t')
        elif '.csv' in file_name_in:
            bits = line.strip().split(',')
        else:
            print('ERROR, only .csv or .tsv files allowed.')
            return()
        # collect the next gene
        if bits[i] in gd:
            gd[ bits[i] ] = gd[ bits[i] ] + [ bits[j] ]   # past items + next gene
        else:
            gd[ bits[i] ] = [ bits[k] ] + [ bits[j] ]  # desc + first gene
    #
    # now we walk the dictionary
    for ki in gd.keys():
        if gs_mode == 'up':
            outtxt.write(ki+'.up'+'\t'+'\t'.join(gd[ki]) +'\n')
        elif gs_mode == 'both':
            outtxt.write(ki+'.both'+'\t'+'\t'.join(gd[ki]) +'\n')
        else:
            outtxt.write(ki+'.dn'+'\t'+'\t'.join(gd[ki]) +'\n')
    return()



from setuptools import setup

def readme():
    with open('README.md') as f:
        return f.read()

setup(name='gssnng',
      version='0.2.6',
      description='Gene Set Scoring on the Nearest Neighbor Graph (gssnng)',
      url='http://github.com/gibbsdavidl/gssnng',
      author='David Gibbs,Michael Strasser',
      author_email='gibbsdavidl@gmail.com',
      license='MIT',
      packages=['gssnng'],
      install_requires=[
          'pandas', 'numpy', 'matplotlib', 'seaborn', 'scipy', 'statsmodels', 'scanpy', 'tqdm'
      ],
      zip_safe=False)
